@@ -19,7 +19,7 @@ try:
 except ImportError:
     from .compat import cached_property
 
-from .context import ctx
+from .context import Context, ctx, get_current_context
 from .exceptions import ParameterDefinitionError, BadArgument, MissingArgument, InvalidChoice, CommandDefinitionError
 from .exceptions import ParamUsageError, ParamConflict, ParamsMissing, NoActiveContext, UnsupportedAction
 from .formatting.utils import HelpEntryFormatter
@@ -56,7 +56,8 @@ Param = TypeVar('Param', bound='Parameter')
 ParamList = List[Param]
 ParamOrGroup = Union[Param, 'ParamGroup']
 
-# TODO: Parameter.validator method that can be used as a decorator?
+# TODO: Parameter.validator method to be used as a decorator, similar to property.setter, replacing type if specified?
+# TODO: constraints / choice=range(1, 101) for example?
 
 
 class parameter_action:
@@ -150,13 +151,31 @@ class ParamBase(ABC):
     def __hash__(self) -> int:
         return reduce(xor, map(hash, (self.__class__, self.__name, self.name, self.command)))
 
+    def _ctx(self, command: 'Command' = None) -> Context:
+        try:
+            return get_current_context()
+        except NoActiveContext:
+            pass
+        command = command or self.command
+        try:
+            return command._Command__ctx
+        except AttributeError:
+            pass
+        raise NoActiveContext('There is no active context')
+
     # region Usage / Help Text
 
     @cached_property
     def formatter(self) -> 'ParamHelpFormatter':
         from .formatting.params import ParamHelpFormatter  # Here due to circular dependency
 
-        return ParamHelpFormatter(self)  # noqa
+        try:
+            with self._ctx() as context:
+                formatter_factory = context.param_formatter or ParamHelpFormatter
+        except NoActiveContext:
+            formatter_factory = ParamHelpFormatter
+
+        return formatter_factory(self)  # noqa
 
     @property
     @abstractmethod
@@ -189,7 +208,7 @@ class ParamGroup(ParamBase):
 
     def __init__(
         self,
-        name: str = None,
+        name: str = None,  # TODO: 'options' always added in help text
         *,
         description: str = None,
         mutually_exclusive: Bool = False,
@@ -460,6 +479,7 @@ class Parameter(ParamBase, ABC):
         :param help: A brief description of this parameter that will appear in ``--help`` text.
         :param hide: If ``True``, this parameter will not be included in usage / help messages.  Defaults to ``False``.
         """
+        # TODO: hide_default: Bool = False
         if action not in self._actions:
             raise ParameterDefinitionError(
                 f'Invalid action={action!r} for {self.__class__.__name__} - valid actions: {sorted(self._actions)}'
@@ -501,11 +521,8 @@ class Parameter(ParamBase, ABC):
         if command is None:
             return self
 
-        try:
+        with self._ctx(command):
             value = self.result()
-        except NoActiveContext:
-            with command._Command__ctx:  # noqa
-                value = self.result()
 
         name = self._name
         if name is not None:
@@ -606,7 +623,8 @@ class Parameter(ParamBase, ABC):
             nargs = self.nargs
             val_count = len(value)
             if val_count == 0 and 0 not in nargs:
-                raise MissingArgument(self)
+                if self.required:
+                    raise MissingArgument(self)
             elif val_count not in nargs:
                 raise BadArgument(self, f'expected nargs={nargs!r} values but found {val_count}')
             elif choices:
@@ -831,7 +849,7 @@ class ChoiceMap(BasePositional):
         if choice_validation_exc is not None:
             cls._choice_validation_exc = choice_validation_exc
 
-    def __init__(self, action: str = 'append', title: str = None, description: str = None, **kwargs):
+    def __init__(self, *, action: str = 'append', title: str = None, description: str = None, **kwargs):
         """
         :param action: The action to take on individual parsed values.  Actions must be defined as methods in classes
           that extend Parameter, and must be registered via :class:`parameter_action`.  Defaults to (and only supports)
@@ -954,16 +972,15 @@ class SubCommand(ChoiceMap, title='Subcommands', choice_validation_exc=CommandDe
     :meth:`.register` sub commands explicitly to specify a different choice value.
     """
 
-    def __init__(self, *args, required: Bool = True, **kwargs):
+    def __init__(self, *, required: Bool = True, **kwargs):
         """
-        :param args: Additional positional arguments to pass to :class:`ChoiceMap`.
         :param required: Whether this parameter is required or not.  If it is required, then an exception will be
           raised if the user did not provide a value for this parameter.  Defaults to ``True``.  If not required and
           not provided, the :meth:`~.commands.Command.main` method for the base :class:`~.commands.Command` that
           contains this SubCommand will be executed by default.
         :param kwargs: Additional keyword arguments to pass to :class:`ChoiceMap`.
         """
-        super().__init__(*args, **kwargs)
+        super().__init__(**kwargs)
         self.required = required
         if not required:
             self._register_choice(None, None)  # Results in next_cmd=None in parse_args, so the base cmd will run
